/*
 * proc_service.c - A minimal subset of the proc_service interface.
 *
 * Copyright: Jan Blunck
 */

#include <stdarg.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>

/* The definitions in this file must correspond to those in the debugger.  */
#include "proc_service.h"
#include "proc_service_impl.h"
#include "symbols.h"

ps_err_e ps_lgetfpregs(struct ps_prochandle *ph, lwpid_t pid,
		       prfpregset_t *regset)
{
	fprintf(stderr, "%s\n", __FUNCTION__);
	return PS_ERR;
}

ps_err_e ps_lsetfpregs(struct ps_prochandle *ph, lwpid_t pid,
		       const prfpregset_t *regset)
{
	fprintf(stderr, "%s\n", __FUNCTION__);
	return PS_ERR;
}

ps_err_e ps_lgetregs(struct ps_prochandle *ph, lwpid_t pid,
		     prgregset_t regset)
{
	fprintf(stderr, "%s\n", __FUNCTION__);
	return ph->ops->lgetregs(ph, pid, regset);
}

ps_err_e ps_lsetregs(struct ps_prochandle *ph, lwpid_t pid,
		     const prgregset_t regset)
{
	fprintf(stderr, "%s\n", __FUNCTION__);
	return PS_ERR;
}

pid_t ps_getpid(struct ps_prochandle *ph)
{
	fprintf(stderr, "%s\n", __FUNCTION__);
	return ph->pid;
}

ps_err_e ps_pglobal_lookup(struct ps_prochandle *ph, const char *object_name,
			   const char *sym_name, psaddr_t *sym_addr)
{
	const Elf64_Addr *sym;

	sym = findLocalSym(sym_name, ph->symtab);

	fprintf(stderr, "%s(%s,%s) = 0x%x\n", __FUNCTION__, object_name,
		sym_name, sym);

	if (!sym)
		return PS_NOSYM;

	*sym_addr = (psaddr_t)sym;
	return PS_OK;
}

ps_err_e ps_pdread(struct ps_prochandle *ph, psaddr_t addr, void *buf,
		   size_t size)
{
	fprintf(stderr, "%s(addr=0x%x,size=%d)\n", __FUNCTION__, addr, size);
	return ph->ops->pdread(ph, addr, buf, size);
}

ps_err_e ps_pdwrite(struct ps_prochandle *ph, psaddr_t addr,
		    const void *buf, size_t size)
{
	fprintf(stderr, "%s\n", __FUNCTION__);
	return PS_ERR;
}

int ps_init(void)
{
	int err;

	err = ps_symbols_init();
	if (err)
		fprintf(stderr, "proc_service init failed\n");

	return err;
}

void ps_exit(void)
{
	ps_symbols_exit();
}

void ps_abort(const char *msg)
{
	ps_exit();
	fprintf(stderr, "%s\n", msg);
	exit(1);
}

void ps_debug(const char* fmt,...)
{
	va_list list;

	va_start(list, fmt);
	fputs("DEBUG: ", stderr);
	vfprintf(stderr, fmt, list);
	va_end(list);
}

void ps_release(struct ps_prochandle *ph)
{
	ph->ops->release(ph);
}

static int read_dynamic_section(struct ps_prochandle *ph)
{
	int err;
	psaddr_t sym;
	ElfW(Dyn) dyn;
	psaddr_t r_debug = 0;
	psaddr_t r_map_addr;
	psaddr_t ldbase_addr;

	err = ps_pglobal_lookup(ph, NULL, "_DYNAMIC", &sym);
	if (err != PS_OK)
		return err;

	do {
		err = ps_pdread(ph, sym, &dyn, sizeof(ElfW(Dyn)));
		if (err != PS_OK)
			return err;

		fprintf(stderr, "dyn: %x 0x%x\n", dyn.d_tag,
			dyn.d_un.d_ptr);

		if (dyn.d_tag == DT_DEBUG) {
			r_debug = (psaddr_t) dyn.d_un.d_ptr;
			fprintf(stderr,"r_debug at 0x%x\n", r_debug);
			break;
		}

		sym += sizeof(ElfW(Dyn));
	} while(dyn.d_tag != DT_NULL);

	if (r_debug == 0)
		return PS_ERR;

	// FIXME: Check the link version and the r_state

	err = ps_pdread(ph, r_debug + offsetof(struct r_debug, r_map),
			&r_map_addr, sizeof(psaddr_t));
	fprintf(stderr,"r_map at 0x%x\n", r_map_addr);
	err = ps_pdread(ph, r_debug + offsetof(struct r_debug, r_ldbase),
			&ldbase_addr, sizeof(psaddr_t));
	fprintf(stderr,"r_ldbase at 0x%x\n", ldbase_addr);

	return 0;
}

#define PATH_MAX 256

struct ps_prochandle *ps_attach(pid_t pid)
{
	int res;
	char filename[PATH_MAX];
	struct ps_prochandle *ph;

	ph = ptrace_attach(pid);
	if (!ph)
		return NULL;

	snprintf(filename, PATH_MAX, "/proc/%d/exe", pid);
	if (res < 0) {
		ps_release(ph);
		return NULL;
	}

	ph->symtab = loadSyms(filename, 0);
	if (!ph->symtab) {
		fprintf(stderr, "No symbols found.\n");
		ps_release(ph);
		return NULL;
	}

	read_dynamic_section(ph);
	return ph;
}
